"""
    Spoc Tools
"""

import dataclasses as dc

from .frozendict import FrozenDict
from .tools import get_attr
from .types import Class, Spoc


def get_spoc(plugins: dict) -> Spoc:
    """Collect All Project Classes"""

    out_dict = {}
    global_dict = {}
    for module_key, module_list in plugins.items():
        out_dict[module_key] = {}
        for current in module_list:
            for current_module, active_class in current.fields.items():
                metadata = get_attr(active_class, "__spoc__")
                if metadata:
                    is_spoc_plugin = get_attr(metadata, "is_spoc_plugin")
                    if is_spoc_plugin:
                        module_uri = f"{current.app}.{current_module.lower()}"
                        global_uri = f"{current.module}.{module_uri}"
                        out_dict[module_key][module_uri] = Class(
                            name=current_module,
                            app=current.app,
                            module=current.module,
                            key=module_uri,
                            uri=global_uri,
                            object=active_class,
                        )
                        global_dict[global_uri] = active_class
        # FrozenDict
        out_dict[module_key] = FrozenDict(**out_dict[module_key])

    # Return Plugin(s) Setup
    dataclass_globals = dc.make_dataclass("Plugin", list(out_dict.keys()), frozen=True)
    app_schema = dataclass_globals(**out_dict)
    return Spoc(schema=app_schema, modules=FrozenDict(**global_dict))



import pathlib
from types import SimpleNamespace

SETTINGS_TEXT = '''
# -*- coding: utf-8 -*-
"""
    { Settings }
"""

import pathlib

# Base Directory
BASE_DIR = pathlib.Path(__file__).parents[1]
'''

CONFIG_TEXT = '''
# -*- coding: utf-8 -*-
"""
    { Config }
"""

from . import settings
'''


SPOC_TEXT = """
[spoc]
mode = "custom" # options: custom, development, production, staging
custom_mode = "development"

# Modes
[spoc.apps]
production = []
development = []
staging = []

# Extras
[spoc.extras]
"""


def create_base():
    import __main__

    root_dir = pathlib.Path(__main__.__file__).parent
    install_helper = SimpleNamespace()

    install_helper.config_dir = root_dir / "config"
    install_helper.config_file = install_helper.config_dir / "__init__.py"
    install_helper.settings_file = install_helper.config_dir / "settings.py"
    install_helper.spoc_file = install_helper.config_dir / "spoc.toml"

    install_helper.config_dir.mkdir(exist_ok=True)

    if not install_helper.config_file.exists():
        with open(install_helper.config_file, "w", encoding="utf-8") as file:
            file.write(CONFIG_TEXT)

    if not install_helper.settings_file.exists():
        with open(install_helper.settings_file, "w", encoding="utf-8") as file:
            file.write(SETTINGS_TEXT)

    if not install_helper.spoc_file.exists():
        with open(install_helper.spoc_file, "w", encoding="utf-8") as file:
            file.write(SPOC_TEXT)



"""
    Exception & Error Messages
"""


class FrozenDict(dict):
    """Immutable Dictionary"""

    def fromkeys(self, S, v):
        """From-Keys"""
        return type(self)(dict(self).fromkeys(S, v))

    def __getattribute__(self, attribute):
        """Get Attribute"""
        if attribute in ("clear", "update", "pop", "popitem", "setdefault"):
            raise AttributeError(
                f"{type(self).__name__} object has no attribute {attribute}"
            )
        return dict.__getattribute__(self, attribute)

    def __setitem__(self, key, value):
        """Set"""
        raise TypeError(
            f"{type(self).__name__} object does not support item assignment"
        )

    def __delitem__(self, key):
        """Delete"""
        raise TypeError(f"{type(self).__name__} object does not support item deletion")

    def __hash__(self):
        """Hash"""
        return hash(tuple(sorted(self.iteritems())))



"""
    Elastic Framework Builder
"""

# Core Tools
from .components import component, is_component

# Frame Work
from .framework import Spoc as App
from .importer import frozendict, get_fields, search
from .singleton import singleton

# App
# extras = Spoc.extras
# plugin = Spoc.plugin

# Globals
base_dir = App.base_dir
config = App.config
mode = App.mode
project = App.project
settings = App.settings



"""
    Object-Class Tools
"""


def get_fields(model: object):
    """Get Object Fields"""
    return [i for i in dir(model) if not i.startswith("__")]


def get_attr(model: object, name: str):
    """Get Object Attribute IF Exist"""
    if hasattr(model, name):
        return getattr(model, name)
    return None



"""
    Auto-Importer
"""

import importlib
import pkgutil

from .frozendict import FrozenDict
from .tools import get_attr, get_fields
from .types import Core, Definition


def iter_namespace(ns_pkg):
    """https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/"""
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


def import_module(single_app: str):
    """Import Single-Module"""
    try:
        module = importlib.import_module(single_app)
    except:
        module = None
    return module


def import_modules(all_apps: list) -> dict:
    """https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/"""
    installed_apps = {}
    for app in all_apps:
        module = import_module(app)
        if module:
            discovered_plugins = {
                name: importlib.import_module(name)
                for finder, name, ispkg in iter_namespace(module)
            }
            installed_apps.update(discovered_plugins)
    return installed_apps


def get_plugins(plugins: list[str], apps: list = None) -> Core:
    """Plugins: Creating & Discovering"""

    plugin_dict = {key: [] for key in plugins}
    installed_apps = import_modules(apps)

    for app_path, module_setup in installed_apps.items():
        uri_parts = app_path.split(".")
        app_name = uri_parts[0]
        app_module = None
        if len(uri_parts) > 1:
            app_module = uri_parts[1]
        if app_module and app_module in plugins:
            current_fields = {}
            for field in get_fields(module_setup):
                current_node = get_attr(module_setup, field)
                if current_node is not None:
                    current_fields[field] = current_node
            plugin = Definition(
                path=app_path,
                app=app_name,
                module=app_module,
                fields=FrozenDict(current_fields),
            )
            plugin_dict[app_module].append(plugin)

    return Core(modules=installed_apps, plugins=plugin_dict)


def search_method(dotted_path: str):
    """Search for Method in <Module>"""
    parts = dotted_path.split(".")
    root = parts[0]
    module = import_module(root)
    import_modules([root])
    for part in parts[1:]:
        module = get_attr(module, part)
    return module



"""
    FrameWork
"""

import os

from .importer import create_framework, frozendict
from .inject import (
    collect_extras,
    collect_installed_apps,
    get_app_mode,
    get_toml_files,
    inject_apps_folder,
    load_envs,
)
from .singleton import singleton


from . import installer

try:
    import config
except ImportError as exception:
    try:
        installer.create_base()
    except:
        config = None

if config:
    try:
        from config import settings
    except ImportError as exception:
        raise ValueError("Missing { ./config/__init__.py } module.") from exception

    if not settings:
        raise ValueError("Missing { ./config/settings.py } module.")

    if not hasattr(settings, "BASE_DIR"):
        raise ValueError(
            """Missing { `BASE_DIR = pathlib.Path(__file__).parents[1]` } in file { ./config/settings.py }."""
        )


    # Core
    PROJECT = config
    SETTINGS = settings
    BASE_DIR = settings.BASE_DIR

    # Global Methods
    inject_apps_folder(BASE_DIR)

    # Global Variables
    TOML_DIR = get_toml_files(BASE_DIR)
    MODE = get_app_mode(TOML_DIR)
    EXTRAS = TOML_DIR["spoc"].get("spoc", {}).get("extras", {})

    # Global Fixed
    TOML_DIR["env"] = load_envs(BASE_DIR, MODE)
    TOML_DIR = frozendict(TOML_DIR)


    @singleton
    class Spoc:
        """Core Object(Singleton) for the whole Framework"""

        base_dir = BASE_DIR
        config = TOML_DIR
        mode = MODE
        project = PROJECT
        settings = SETTINGS

        def init(self, plugins: list = None):
            """Finally: Collect { Keys }"""
            # GLOBALS
            installed_apps = collect_installed_apps(TOML_DIR, SETTINGS)
            extras = collect_extras(EXTRAS)

            plugins = plugins or []
            framework = create_framework(
                plugins=plugins, installed_apps=installed_apps, extras=extras
            )

            # Output Model
            self.installed_apps = installed_apps
            self.extras = framework.extras
            self.module = framework.module
            self.component = framework.plugin

            os.chdir(BASE_DIR)

        @staticmethod
        def keys():
            """Object Keys"""
            return [
                "base_dir",
                "config",
                "installed_apps",
                "mode",
                "project",
                "settings",
                "extras",
                "module",
                "component",
            ]



from framework import commands, components

import spoc

@commands
def test():
    print("Hello World")


is_my_type = spoc.is_component(test, components["command"])

print(f"""
Hello From 'apps/demo/commands.py'
is the plugin of the <type> I designed? {is_my_type}      
""")



"""[ Collect & Inject ]
    Project (Module)
"""

import os
import pathlib
import sys

from .importer import frozendict, search
from .toml_core import TOML


def inject_apps_folder(base_dir=None):
    """Inject { ./apps } Directory"""

    base_apps = pathlib.Path(base_dir / "apps")
    base_apps.mkdir(parents=True, exist_ok=True)
    if base_apps.exists():
        sys.path.insert(0, os.path.join(base_dir, "apps"))

def collect_apps_partial(app_mode, apps):
    installed_apps = []
    match app_mode:
        case "production":
            installed_apps.extend(apps.get("production", []))
        case "staging":  # production + staging
            installed_apps.extend(apps.get("production", []))
            installed_apps.extend(apps.get("staging", []))
        case "development":  # production + staging + development
            installed_apps.extend(apps.get("production", []))
            installed_apps.extend(apps.get("staging", []))
            installed_apps.extend(apps.get("development", []))
    return installed_apps


def collect_installed_apps(toml_dir, settings=None):
    """Collect All Apps"""

    # Step[1]: INIT { Values }
    toml_spoc = toml_dir["spoc"].get("spoc", {})
    app_mode = toml_spoc.get("mode", "development")
    custom_mode = toml_spoc.get("custom_mode", None)
    apps = toml_spoc.get("apps", {})
    installed_apps = []

    # Step[2]: Collect Apps
    match app_mode:
        case "custom":
            installed_apps.extend(collect_apps_partial(custom_mode, apps))
            if hasattr(settings, "INSTALLED_APPS"):
                installed_apps.extend(settings.INSTALLED_APPS)
        case _:
            installed_apps.extend(collect_apps_partial(app_mode, apps))

    return installed_apps


def get_toml_file(file: str):
    """Get { TOML } File"""

    toml_file = pathlib.Path(file)
    manager = TOML(toml_file)

    if toml_file.exists():
        toml_dict = frozendict(**manager.read())
    else:
        toml_dict = frozendict({})
    return toml_dict


def get_toml_files(base_dir=None):
    """Collect All { TOML } Files"""

    return {
        "spoc": get_toml_file(base_dir / "config" / "spoc.toml"),
        "pyproject": get_toml_file(base_dir / "pyproject.toml"),
    }


def get_app_mode(toml_dir):
    """Get App { Mode }"""

    mode = toml_dir["spoc"].get("mode", "development")
    if mode == "custom":
        mode = toml_dir["spoc"].get("custom_mode", "development")
    return mode


def collect_extras(extras):
    """Collect All Extra { Python-Objects }"""

    all_extras_modules = {}
    for group, items in extras.items():
        all_extras_modules[group] = []
        for object_uri in items:
            obj = search(object_uri)
            if obj:
                all_extras_modules[group].append(obj)
    return frozendict(all_extras_modules)


def load_envs(base_dir, mode):
    """Load Environment Variables"""
    file_path = base_dir / "config" / ".env" / f"{ mode }.toml"
    return get_toml_file(file_path).get("env", {})



"""
    Components
"""
import spoc

components = {}
components["command"] = {"type": "command"}

# Class @Decorator
def commands(
    cls: object = None,
):
    """Demo"""
    spoc.component(cls, metadata=components["command"])
    return cls


"""
    Tool for TOML (before tomlib gets implemented)
"""

import toml


class TOML:
    """TOML Wrapper"""

    def __init__(self, file=None):
        """Initialize TOML-Manager"""

        self.file = file

        if not file:
            raise ValueError("Please Set a <Path> for the <TOML> file.")

    def read(self):
        """Read"""

        with open(self.file, "r", encoding="utf-8") as active_file:
            toml_string = active_file.read()
            parsed_toml = toml.loads(toml_string)
        return parsed_toml



from framework import MyFramework
import spoc 

test = MyFramework()

print(test.env)

print(test.component.commands['demo.test'].object)

for method in test.extras['middleware']:
    print(method)
    

# print(test.plugin.commands.values())
print(spoc.settings)



"""
Framework
"""

import spoc

PLUGINS = ["commands", "models", "views"]

@spoc.singleton
class MyFramework:
    """Framework"""

    def init(
        self,
    ):
        """Class __init__ Replacement"""
        framework = spoc.App(plugins=PLUGINS)
                
        self.base_dir=framework.base_dir
        self.mode = framework.mode
        self.env=framework.config["env"]
        self.pyproject=framework.config["pyproject"]
        self.spoc=framework.config["spoc"]
        self.settings = framework.settings        
        self.component = framework.component
        self.extras = framework.extras
                
        self.keys = [
            "base_dir",
            "mode",
            "env",
            "pyproject",
            "spoc",
            "settings",
            "component",
            "extras",            
        ]


# -*- coding: utf-8 -*-
""" [Settings]
    Project Settings { app.settings }.
"""
import pathlib

# Base Directory
BASE_DIR = pathlib.Path(__file__).parents[1]

# Installed Apps
INSTALLED_APPS = ["demo"]



"""
    Components
"""

import dataclasses as dc
import functools
import typing


@dc.dataclass(frozen=True)
class Component:
    """Spoc Plugin"""

    config: typing.Any = None
    metadata: typing.Any = None
    is_spoc_plugin: bool = True


def component(
    cls: object = None,
    *,
    config: dict = None,
    metadata: dict = None,
) -> typing.Any:
    """Plugin Creator"""

    config = config or {}
    metadata = metadata or {}
    if cls is None:
        return functools.partial(
            component,
            config=config,
            metadata=metadata,
        )

    # Real Wrapper
    cls.__spoc__ = Component(config=config, metadata=metadata)
    return cls


def is_component(cls, metadata: dict = None):
    """Plugin Validator"""
    return cls.__spoc__.metadata == metadata
